import pytest

# fmt: off
from icupy.icu import (
    U_ICU_VERSION_MAJOR_NUM, Locale, LocaleDisplayNames, UDialectHandling,
    UnicodeString, UScriptCode,
)

# fmt: on


def test_api():
    ldn = LocaleDisplayNames.create_instance(
        Locale.get_us(), UDialectHandling.ULDN_DIALECT_NAMES
    )

    # UDialectHandling icu::LocaleDisplayNames::getDialectHandling()
    assert ldn.get_dialect_handling() == UDialectHandling.ULDN_DIALECT_NAMES

    # const Locale &icu::LocaleDisplayNames::getLocale()
    result = ldn.get_locale()
    assert isinstance(result, Locale)
    assert result == Locale("en-US")

    temp = UnicodeString()

    # UnicodeString &icu::LocaleDisplayNames::keyDisplayName(
    #       const char *key,
    #       UnicodeString &result
    # )
    result = ldn.key_display_name("efg", temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "efg"

    # UnicodeString &icu::LocaleDisplayNames::keyValueDisplayName(
    #       const char *key,
    #       const char *value,
    #       UnicodeString &result
    # )
    result = ldn.key_value_display_name("ca", "ijk", temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "ijk"

    # UnicodeString &icu::LocaleDisplayNames::languageDisplayName(
    #       const char *lang,
    #       UnicodeString &result
    # )
    result = ldn.language_display_name("xy", temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "xy"

    # [1]
    # UnicodeString &icu::LocaleDisplayNames::localeDisplayName(
    #       const char *localeId,
    #       UnicodeString &result
    # )
    result = ldn.locale_display_name("xy", temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "xy"

    # [2]
    # UnicodeString &icu::LocaleDisplayNames::localeDisplayName(
    #       const Locale &locale,
    #       UnicodeString &result
    # )
    result = ldn.locale_display_name(Locale("xy"), temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "xy"

    # UnicodeString &icu::LocaleDisplayNames::regionDisplayName(
    #       const char *region,
    #       UnicodeString &result
    # )
    result = ldn.region_display_name("xy-Latn-wx", temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "xy-Latn-wx"

    # [1]
    # UnicodeString &icu::LocaleDisplayNames::scriptDisplayName(
    #       const char *script,
    #       UnicodeString &result
    # )
    result = ldn.script_display_name("wxyz", temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "wxyz"

    # [2]
    # UnicodeString &icu::LocaleDisplayNames::scriptDisplayName(
    #       UScriptCode scriptCode,
    #       UnicodeString &result
    # )
    result = ldn.script_display_name(UScriptCode.USCRIPT_LATIN, temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "Latin"

    # UnicodeString &icu::LocaleDisplayNames::variantDisplayName(
    #       const char *variant,
    #       UnicodeString &result
    # )
    result = ldn.variant_display_name("abc", temp)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(temp)
    assert result == "abc"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 51, reason="ICU4C<51")
def test_api_51():
    from icupy.icu import UDisplayContext, UDisplayContextType

    ldn = LocaleDisplayNames.create_instance(Locale.get_germany())

    # UDisplayContext icu::LocaleDisplayNames::getContext(
    #       UDisplayContextType type
    # )
    result = ldn.get_context(
        UDisplayContextType.UDISPCTX_TYPE_SUBSTITUTE_HANDLING
    )
    assert result == UDisplayContext.UDISPCTX_SUBSTITUTE


def test_create_instance():
    temp = UnicodeString()

    # [1]
    # static LocaleDisplayNames *icu::LocaleDisplayNames::createInstance(
    #       const Locale &locale
    # )
    ldn = LocaleDisplayNames.create_instance(Locale.get_germany())
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp == "xy"
    ldn.locale_display_name("en", temp)
    assert temp == "Englisch"

    ldn = LocaleDisplayNames.create_instance("de-DE")
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp == "xy"
    ldn.locale_display_name("en", temp)
    assert temp == "Englisch"

    # [2]
    # static LocaleDisplayNames *icu::LocaleDisplayNames::createInstance(
    #       const Locale &locale,
    #       UDialectHandling dialectHandling
    # )
    ldn = LocaleDisplayNames.create_instance(
        Locale.get_us(), UDialectHandling.ULDN_DIALECT_NAMES
    )
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp == "xy"
    ldn.locale_display_name("en", temp)
    assert temp == "English"

    ldn = LocaleDisplayNames.create_instance(
        "en-US", UDialectHandling.ULDN_DIALECT_NAMES
    )
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp == "xy"
    ldn.locale_display_name("en", temp)
    assert temp == "English"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 51, reason="ICU4C<51")
def test_create_instance_51():
    from icupy.icu import UDisplayContext

    temp = UnicodeString()

    # [3]
    # static LocaleDisplayNames *icu::LocaleDisplayNames::createInstance(
    #       const Locale &locale,
    #       UDisplayContext *contexts,
    #       int32_t length
    # )
    contexts = [
        UDisplayContext.UDISPCTX_NO_SUBSTITUTE,
        UDisplayContext.UDISPCTX_DIALECT_NAMES,
    ]
    ldn = LocaleDisplayNames.create_instance(
        Locale.get_germany(), contexts, len(contexts)
    )
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp.is_bogus()
    ldn.locale_display_name("en", temp)
    assert temp == "Englisch"

    ldn = LocaleDisplayNames.create_instance("de-DE", contexts, len(contexts))
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp.is_bogus()
    ldn.locale_display_name("en", temp)
    assert temp == "Englisch"

    # length is optional
    ldn = LocaleDisplayNames.create_instance(Locale.get_germany(), contexts)
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp.is_bogus()
    ldn.locale_display_name("en", temp)
    assert temp == "Englisch"

    ldn = LocaleDisplayNames.create_instance("de-DE", contexts)
    assert isinstance(ldn, LocaleDisplayNames)
    ldn.locale_display_name("xy", temp)
    assert temp.is_bogus()
    ldn.locale_display_name("en", temp)
    assert temp == "Englisch"



import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 50:
    pytest.skip("ICU4C<50", allow_module_level=True)

from icupy.icu import GenderInfo, Locale, UGender


def test_api():
    # static const GenderInfo *icu::GenderInfo::getInstance(
    #       const Locale &locale,
    #       UErrorCode &status
    # )
    gi = GenderInfo.get_instance(Locale.get_us())
    assert isinstance(gi, GenderInfo)

    # UGender icu::GenderInfo::getListGender(
    #       const UGender *genders,
    #       int32_t length,
    #       UErrorCode &status
    # )
    assert gi.get_list_gender([], 0) == UGender.UGENDER_OTHER

    assert (
        gi.get_list_gender([UGender.UGENDER_MALE], 1) == UGender.UGENDER_MALE
    )

    assert (
        gi.get_list_gender([UGender.UGENDER_FEMALE, UGender.UGENDER_MALE], 2)
        == UGender.UGENDER_OTHER
    )

    gi = GenderInfo.get_instance("fr")
    assert isinstance(gi, GenderInfo)

    # length is optional
    assert gi.get_list_gender([]) == UGender.UGENDER_OTHER

    assert gi.get_list_gender([UGender.UGENDER_MALE]) == UGender.UGENDER_MALE

    assert (
        gi.get_list_gender([UGender.UGENDER_FEMALE, UGender.UGENDER_MALE])
        == UGender.UGENDER_MALE
    )



# fmt: off
from icupy.icu import (
    UDateTimeScale, UTimeScaleValue, utmscale_from_int64,
    utmscale_get_time_scale_value, utmscale_to_int64,
)

# fmt: on


def test_api():
    # int64_t utmscale_getTimeScaleValue(
    #       UDateTimeScale timeScale,
    #       UTimeScaleValue value,
    #       UErrorCode *status
    # )
    from_min = utmscale_get_time_scale_value(
        UDateTimeScale.UDTS_JAVA_TIME, UTimeScaleValue.UTSV_FROM_MIN_VALUE
    )
    assert isinstance(from_min, int)

    # int64_t utmscale_fromInt64(
    #       int64_t otherTime,
    #       UDateTimeScale timeScale,
    #       UErrorCode *status
    # )
    universal_time = utmscale_from_int64(
        from_min, UDateTimeScale.UDTS_JAVA_TIME
    )
    assert isinstance(universal_time, int)

    # int64_t utmscale_toInt64(
    #       int64_t universalTime,
    #       UDateTimeScale timeScale,
    #       UErrorCode *status
    # )
    dt = utmscale_to_int64(universal_time, UDateTimeScale.UDTS_JAVA_TIME)
    assert isinstance(dt, int)
    assert dt == from_min



import os
import subprocess
import sys

from setuptools import Extension
from setuptools.command.build_ext import build_ext

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name))
        )
        extdir = os.path.join(extdir, ext.name)

        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        env = os.environ.copy()
        cmake_build_type = env.get("CMAKE_BUILD_TYPE", "Release")
        cfg = "Debug" if self.debug else cmake_build_type

        env["CXXFLAGS"] = " ".join(
            [
                env.get("CXXFLAGS", "").strip(),
                '-DVERSION_INFO=\\"{}\\"'.format(
                    self.distribution.get_version()
                ),
            ]
        ).strip()

        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
            "-DCMAKE_BUILD_TYPE=" + cfg,
        ]

        build_args = []

        if sys.platform.startswith("win"):
            if "ICU_ROOT" not in env:
                env["ICU_ROOT"] = "C:\\icu"
                self.warn(
                    "ICU_ROOT environment variable was not set. "
                    "Using default '{}'.".format(env["ICU_ROOT"])
                )
            cmake_args += [
                "-Wno-dev",
            ]

        cmake_generator = env.get("CMAKE_GENERATOR")
        is_msbuild = self.compiler.compiler_type == "msvc" and (
            cmake_generator is None
            or cmake_generator.startswith("Visual Studio")
        )
        is_multi_config = is_msbuild or cmake_generator in [
            "Ninja Multi-Config",
            "Xcode",
        ]

        if is_multi_config:
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(
                    cfg.upper(), extdir
                ),
            ]
            build_args += [
                "--config",
                cfg,
            ]

        if is_msbuild:
            if "CMAKE_GENERATOR_PLATFORM" not in env:
                cmake_args += [
                    "-A",
                    PLAT_TO_CMAKE[self.plat_name],
                ]

        if (
            "CMAKE_BUILD_PARALLEL_LEVEL" not in env
            and not is_msbuild
            and (
                cmake_generator is None
                or not cmake_generator.startswith("Ninja")
            )
        ):
            jobs = (
                hasattr(self, "parallel") and self.parallel
            ) or os.cpu_count()
            if jobs:
                build_args += [
                    "-j",
                    str(jobs),
                ]

        self.announce(
            "-- CXXFLAGS environment variable: " + repr(env.get("CXXFLAGS")),
            level=2,
        )
        self.announce(
            "-- CMake environment variables: "
            + repr(
                [
                    "{}={}".format(k, v)
                    for k, v in env.items()
                    if k.upper().startswith("CMAKE")
                ]
            ),
            level=2,
        )
        self.announce(
            "-- CMake build system options: " + repr(cmake_args), level=2
        )
        self.announce("-- CMake build options: " + repr(build_args), level=2)

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args,
            cwd=self.build_temp,
            env=env,
        )


def build(setup_kwargs):
    src = (
        setup_kwargs["package_dir"][""]
        if "package_dir" in setup_kwargs
        else ""
    )
    ext_modules = [CMakeExtension(setup_kwargs["name"], src)]
    cmdclass = dict(build_ext=CMakeBuild)
    setup_kwargs.update(
        {
            "ext_modules": ext_modules,
            "cmdclass": cmdclass,
        }
    )



import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 51:
    pytest.skip("ICU4C<51", allow_module_level=True)

from icupy.icu import Region, StringEnumeration, URegionType


def test_api():
    world = Region.get_instance("001")
    continent = Region.get_instance("039")  # Southern Europe
    territory = Region.get_instance("IT")  # Italy

    # UBool icu::Region::contains(const Region &other)
    assert not territory.contains(continent)
    assert continent.contains(territory)

    # __contains__(self, item: Region)
    assert continent not in territory
    assert territory in continent

    # [1]
    # const Region *icu::Region::getContainingRegion()
    result = world.get_containing_region()
    assert result is None

    result = territory.get_containing_region()
    assert isinstance(result, Region)
    assert result.contains(territory)

    # [2]
    # const Region *icu::Region::getContainingRegion(URegionType type)
    result = world.get_containing_region(URegionType.URGN_CONTINENT)
    assert result is None

    result = territory.get_containing_region(URegionType.URGN_CONTINENT)
    assert isinstance(result, Region)

    # int32_t icu::Region::getNumericCode()
    assert world.get_numeric_code() == 1
    assert continent.get_numeric_code() == 39
    assert territory.get_numeric_code() == 380

    # const char *icu::Region::getRegionCode()
    result = world.get_region_code()
    assert isinstance(result, str)
    assert result == "001"

    result = continent.get_region_code()
    assert isinstance(result, str)
    assert result == "039"

    result = territory.get_region_code()
    assert isinstance(result, str)
    assert result == "IT"

    # URegionType icu::Region::getType()
    assert world.get_type() == URegionType.URGN_WORLD
    assert continent.get_type() == URegionType.URGN_SUBCONTINENT
    assert territory.get_type() == URegionType.URGN_TERRITORY

    # bool icu::Region::operator!=(const Region &that)
    region = Region.get_instance(380)
    assert continent != territory
    assert continent != region
    assert not (territory != region)

    # bool icu::Region::operator==(const Region &that)
    assert not (continent == territory)
    assert not (continent == region)
    assert territory == region


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 55, reason="ICU4C<55")
def test_api_55():
    # static StringEnumeration *icu::Region::getAvailable(
    #       URegionType type,
    #       UErrorCode &status
    # )
    territories = Region.get_available(URegionType.URGN_GROUPING)
    assert isinstance(territories, StringEnumeration)
    assert len(territories) > 0

    world = Region.get_instance("001")

    # [1]
    # StringEnumeration *icu::Region::getContainedRegions(UErrorCode &status)
    territories = world.get_contained_regions()
    assert isinstance(territories, StringEnumeration)
    assert len(territories) > 0

    # [2]
    # StringEnumeration *icu::Region::getContainedRegions(
    #       URegionType type,
    #       UErrorCode &status
    # )
    territories = world.get_contained_regions(URegionType.URGN_TERRITORY)
    assert isinstance(territories, StringEnumeration)
    assert len(territories) > 0

    # StringEnumeration *icu::Region::getPreferredValues(UErrorCode &status)
    territories = world.get_preferred_values()
    assert territories is None

    grouping = Region.get_instance("SU")
    territories = grouping.get_preferred_values()
    assert isinstance(territories, StringEnumeration)
    assert len(territories) > 0


def test_get_instance():
    # [1]
    # static const Region *icu::Region::getInstance(
    #       const char *region_code,
    #       UErrorCode &status
    # )
    region = Region.get_instance("EU")
    assert isinstance(region, Region)

    # [2]
    # static const Region *icu::Region::getInstance(
    #       int32_t code,
    #       UErrorCode &status
    # )
    region = Region.get_instance(967)
    assert isinstance(region, Region)



# fmt: off
from icupy.icu import (
    U_SHAPE_DIGIT_TYPE_AN, U_SHAPE_DIGIT_TYPE_AN_EXTENDED,
    U_SHAPE_DIGITS_AN2EN, U_SHAPE_DIGITS_EN2AN, u_shape_arabic,
)

# fmt: on


def test_api():
    # From icu/source/test/cintltst/cbiditst.c
    source = "\x31\u0627\x32\u06f3\x61\x34"

    # int32_t u_shapeArabic(
    #       const UChar *source,
    #       int32_t sourceLength,
    #       UChar *dest,
    #       int32_t destSize,
    #       uint32_t options,
    #       UErrorCode *pErrorCode
    # )
    # european -> arabic
    options = U_SHAPE_DIGITS_EN2AN | U_SHAPE_DIGIT_TYPE_AN
    dest = u_shape_arabic(source, len(source), options)
    assert isinstance(dest, str)
    assert dest == "\u0661\u0627\u0662\u06f3\x61\u0664"

    # arabic -> european
    options = U_SHAPE_DIGITS_AN2EN | U_SHAPE_DIGIT_TYPE_AN_EXTENDED
    dest = u_shape_arabic(source, -1, options)
    assert isinstance(dest, str)
    assert dest == "\x31\u0627\x32\x33\x61\x34"



from pathlib import Path

import pytest

# fmt: off
from icupy.icu import (
    USPREP_ALLOW_UNASSIGNED, USPREP_DEFAULT, ICUError, UErrorCode, UParseError,
    UStringPrepProfileType, u_unescape, usprep_close, usprep_open,
    usprep_open_by_type, usprep_prepare,
)
from icupy.utils import gc

# fmt: on


def test_open():
    path = Path(__file__).resolve().parent / "testdata"
    try:
        # UStringPrepProfile *usprep_open(
        #       const char *path,
        #       const char *fileName,
        #       UErrorCode *status
        # )
        #
        # void usprep_close(UStringPrepProfile *profile)
        with gc(usprep_open(str(path), "nfscis"), usprep_close) as prep:
            _ = usprep_prepare(prep, "", 0, USPREP_DEFAULT, None)
    except ICUError as ex:
        if ex.args[0] != UErrorCode.U_FILE_ACCESS_ERROR:
            raise
        pytest.skip(
            "testdata.dat is not found (not an error). "
            "You need to build a test data from the source. "
            "See also <icu4c>/icu/source/test/testdata/"
        )


def test_open_by_type():
    # From icu/source/test/cintltst/spreptst.c

    # UStringPrepProfile *usprep_openByType(
    #       UStringPrepProfileType type,
    #       UErrorCode *status
    # )
    #
    # void usprep_close(UStringPrepProfile *profile)
    with gc(
        usprep_open_by_type(UStringPrepProfileType.USPREP_RFC4013_SASLPREP),
        usprep_close,
    ) as prep:
        # int32_t usprep_prepare(
        #       const UStringPrepProfile *prep,
        #       const UChar *src,
        #       int32_t srcLength,
        #       UChar *dest,
        #       int32_t destCapacity,
        #       int32_t options,
        #       UParseError *parseError,
        #       UErrorCode *status
        # )
        src = u_unescape("user:\\u00A0\\u0AC6\\u1680\\u00ADpassword1")
        expected = u_unescape("user: \\u0AC6 password1")
        parse_error = UParseError()
        dest = usprep_prepare(
            prep, src, len(src), USPREP_ALLOW_UNASSIGNED, parse_error
        )
        assert isinstance(dest, str)
        assert dest == expected

    with gc(
        usprep_open_by_type(UStringPrepProfileType.USPREP_RFC4011_MIB),
        usprep_close,
    ) as prep:
        src = u_unescape("Policy\\u034F\\u200DBase\\u0020d\\u1806\\u200C")
        expected = u_unescape("PolicyBase d")
        dest = usprep_prepare(prep, src, -1, USPREP_ALLOW_UNASSIGNED, None)
        assert isinstance(dest, str)
        assert dest == expected



from functools import partial

import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 59:
    pytest.skip("ICU4C<59", allow_module_level=True)

from icupy.icu import Edits, ErrorCode, UErrorCode


# From icu/source/test/intltest/strcase.cpp
def test_api_59():
    # [1]
    # icu::Edits::Edits()
    edits = Edits()

    # void icu::Edits::add_replace(
    #       int32_t oldLength,
    #       int32_t newLength
    # )
    edits.add_replace(2, 1)
    edits.add_replace(2, 1)
    edits.add_replace(2, 1)
    edits.add_replace(0, 10)
    edits.add_replace(100, 0)
    edits.add_replace(3000, 4000)
    edits.add_replace(100000, 100000)

    # void icu::Edits::add_unchanged(int32_t unchangedLength)
    edits.add_unchanged(1)
    edits.add_unchanged(10000)
    edits.add_unchanged(2)
    edits.add_unchanged(0)

    # UBool icu::Edits::copyErrorTo(UErrorCode &outErrorCode)
    out_error_code = ErrorCode()
    assert not edits.copy_error_to(out_error_code)
    assert out_error_code == UErrorCode.U_ZERO_ERROR

    out_error_code.set(UErrorCode.U_ILLEGAL_ARGUMENT_ERROR)
    assert edits.copy_error_to(out_error_code)
    assert out_error_code == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR

    # UBool icu::Edits::hasChanges()
    assert edits.has_changes()

    # int32_t icu::Edits::lengthDelta()
    assert edits.length_delta() == -3 + 10 - 100 + 1000

    # Iterator icu::Edits::getCoarseChangesIterator()
    ei = edits.get_coarse_changes_iterator()
    assert isinstance(ei, Edits.Iterator)

    # Iterator icu::Edits::getCoarseIterator()
    ei = edits.get_coarse_iterator()
    assert isinstance(ei, Edits.Iterator)

    # Iterator icu::Edits::getFineChangesIterator()
    ei = edits.get_fine_changes_iterator()
    assert isinstance(ei, Edits.Iterator)

    # Iterator icu::Edits::getFineIterator()
    ei = edits.get_fine_iterator()
    assert isinstance(ei, Edits.Iterator)

    # void icu::Edits::reset()
    edits.reset()
    assert not edits.has_changes()


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 60, reason="ICU4C<60")
def test_api_60():
    other = Edits()
    other.add_unchanged(1)
    other.add_unchanged(10000)
    other.add_replace(0, 0)
    other.add_unchanged(2)
    other.add_replace(2, 1)
    other.add_replace(2, 1)
    other.add_replace(2, 1)
    other.add_replace(0, 10)
    other.add_replace(100, 0)
    other.add_replace(3000, 4000)
    other.add_replace(100000, 100000)

    # [2]
    # icu::Edits::Edits(const Edits &other)
    edits = Edits(other)

    # int32_t icu::Edits::numberOfChanges()
    assert edits.number_of_changes() == 7

    # Edits &icu::Edits::mergeAndAppend(
    #       const Edits &ab,
    #       const Edits &bc,
    #       UErrorCode &errorCode
    # )
    edits.reset()
    ab = Edits()
    bc = Edits()
    expected = Edits()

    ab.add_unchanged(2)
    bc.add_unchanged(2)
    expected.add_unchanged(2)

    ab.add_replace(3, 2)
    bc.add_replace(2, 1)
    expected.add_replace(3, 1)

    ab.add_unchanged(5)
    bc.add_unchanged(3)
    expected.add_unchanged(3)

    ab.add_replace(4, 3)
    bc.add_replace(3, 2)
    ab.add_replace(4, 3)
    bc.add_replace(3, 2)
    ab.add_replace(4, 3)
    bc.add_replace(3, 2)
    bc.add_unchanged(4)
    expected.add_replace(14, 8)

    ab.add_unchanged(2)
    expected.add_unchanged(2)

    ab.add_replace(0, 5)
    ab.add_replace(0, 2)
    bc.add_replace(7, 0)

    ab.add_replace(1, 2)
    bc.add_replace(2, 3)
    expected.add_replace(1, 3)

    ab.add_replace(1, 0)
    ab.add_replace(2, 0)
    ab.add_replace(3, 0)
    expected.add_replace(1, 0)
    expected.add_replace(2, 0)
    expected.add_replace(3, 0)

    ab.add_unchanged(2)
    bc.add_unchanged(1)
    expected.add_unchanged(1)

    bc.add_replace(0, 4)
    bc.add_replace(0, 5)
    bc.add_replace(0, 6)
    expected.add_replace(0, 4)
    expected.add_replace(0, 5)
    expected.add_replace(0, 6)

    bc.add_replace(2, 2)
    ab.add_replace(1, 0)
    ab.add_replace(2, 0)
    ab.add_replace(3, 0)
    ab.add_replace(4, 1)
    expected.add_replace(11, 2)

    ab.add_replace(5, 6)
    bc.add_replace(3, 3)
    bc.add_replace(0, 4)
    bc.add_replace(0, 5)
    bc.add_replace(0, 6)
    bc.add_replace(3, 7)
    expected.add_replace(5, 25)

    ab.add_replace(4, 4)
    ab.add_replace(3, 0)
    ab.add_unchanged(2)
    bc.add_replace(2, 2)
    bc.add_replace(4, 0)
    expected.add_replace(9, 2)

    ab.add_replace(0, 2)
    bc.add_replace(1, 1)
    bc.add_replace(0, 8)
    bc.add_unchanged(4)
    expected.add_replace(0, 10)

    ab.add_unchanged(3)
    expected.add_unchanged(3)

    ab.add_replace(2, 0)
    ab.add_replace(4, 0)
    ab.add_replace(6, 0)
    bc.add_replace(0, 1)
    bc.add_replace(0, 3)
    bc.add_replace(0, 5)
    expected.add_replace(0, 1)
    expected.add_replace(0, 3)
    expected.add_replace(0, 5)
    expected.add_replace(2, 0)
    expected.add_replace(4, 0)
    expected.add_replace(6, 0)

    ab.add_unchanged(1)
    bc.add_unchanged(1)
    expected.add_unchanged(1)

    result = edits.merge_and_append(ab, bc)
    assert isinstance(result, Edits)
    assert id(result) == id(edits)

    ei1 = edits.get_fine_iterator()
    ei2 = expected.get_fine_iterator()
    t1 = list(iter(partial(ei1.next), False))
    t2 = list(iter(partial(ei2.next), False))
    assert t1 == t2


def test_edits_iterator_59():
    edits = Edits()
    edits.add_unchanged(1)
    edits.add_unchanged(10000)
    edits.add_replace(0, 0)
    edits.add_unchanged(2)
    edits.add_replace(2, 1)
    edits.add_unchanged(0)
    edits.add_replace(2, 1)
    edits.add_replace(2, 1)
    edits.add_replace(0, 10)
    edits.add_replace(100, 0)
    edits.add_replace(3000, 4000)
    edits.add_replace(100000, 100000)

    # Iterator icu::Edits::getCoarseChangesIterator()
    ei = edits.get_coarse_changes_iterator()
    assert isinstance(ei, Edits.Iterator)

    exp_src_index = exp_dest_index = exp_repl_index = 0
    exp_old_length = 10003
    exp_new_length = 10003

    # int32_t icu::Edits::Iterator::destinationIndex()
    assert ei.destination_index() == exp_dest_index

    # UBool icu::Edits::Iterator::findSourceIndex(
    #       int32_t i,
    #       UErrorCode &errorCode
    # )
    assert ei.find_source_index(exp_src_index)

    # UBool icu::Edits::Iterator::hasChange()
    assert not ei.has_change()

    # int32_t icu::Edits::Iterator::newLength()
    assert ei.new_length() == exp_new_length

    # int32_t icu::Edits::Iterator::oldLength()
    assert ei.old_length() == exp_old_length

    # int32_t icu::Edits::Iterator::replacementIndex()
    assert ei.replacement_index() == exp_repl_index

    # int32_t icu::Edits::Iterator::sourceIndex()
    assert ei.source_index() == exp_src_index

    # UBool icu::Edits::Iterator::next(UErrorCode &errorCode)
    assert ei.next()
    exp_src_index += exp_old_length
    exp_dest_index += exp_new_length
    exp_old_length = 103106
    exp_new_length = 104013

    assert ei.destination_index() == exp_dest_index
    assert ei.has_change()
    assert ei.new_length() == exp_new_length
    assert ei.old_length() == exp_old_length
    assert ei.replacement_index() == exp_repl_index
    assert ei.source_index() == exp_src_index

    assert not ei.next()


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 60, reason="ICU4C<60")
def test_edits_iterator_60():
    from icupy.icu import CaseMap

    src = "abcßDeF"
    edits = Edits()
    CaseMap.fold(0, src, -1, edits)
    ei = edits.get_fine_iterator()

    # int32_t icu::Edits::Iterator::destinationIndexFromSourceIndex(
    #       int32_t i,
    #       UErrorCode &errorCode
    # )
    indices = []
    for i in range(len(src)):
        assert ei.find_source_index(i)
        indices.append(ei.destination_index_from_source_index(i))
    assert indices == [0, 1, 2, 3, 5, 6, 7]

    # UBool icu::Edits::Iterator::findDestinationIndex(
    #       int32_t i,
    #       UErrorCode &errorCode
    # )
    indices = []
    for i in range(len(src)):
        assert ei.find_destination_index(i)
        indices.append(ei.source_index())
    assert indices == [0, 0, 0, 3, 3, 4, 5]

    # int32_t icu::Edits::Iterator::sourceIndexFromDestinationIndex(
    #       int32_t i,
    #       UErrorCode &errorCode
    # )
    indices = []
    for i in range(len(src)):
        assert ei.find_destination_index(i)
        indices.append(ei.source_index_from_destination_index(i))
    assert indices == [0, 1, 2, 3, 4, 4, 5]



"""
Python bindings for ICU4C
"""

import sys

if sys.platform.startswith("win"):
    import os
    from pathlib import Path

    path = (  # type: ignore
        Path(os.getenv("ICU_ROOT", "C:/icu")) / "bin64"
        if sys.maxsize > 2**32
        else "bin"
    ).resolve()
    if not path.is_dir():
        raise FileNotFoundError(
            "%s is not a valid directory. "
            "Check the ICU_ROOT environment variable setting (ICU_ROOT=%s)"
            % (path, os.getenv("ICU_ROOT", ""))
        )
    if sys.version_info[:2] >= (3, 8):
        os.add_dll_directory(path)
    else:
        os.environ["PATH"] = "{}{}{}".format(
            path, os.pathsep, os.getenv("PATH", "")
        )
    del path

from . import icu  # type: ignore

__version__ = icu.VERSION_INFO

# DEPRECATED: ICUException is deprecated. Use ICUError instead.
# TODO: Remove ICUException in a future release.
ICUException = icu.ICUError



import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 56:
    pytest.skip("ICU4C<56", allow_module_level=True)

import copy

# fmt: off
from icupy.icu import (
    U_ICU_VERSION_MAJOR_NUM, BreakIterator, CharacterIterator,
    FilteredBreakIteratorBuilder, Locale, StringCharacterIterator,
    ULocDataLocaleType, UnicodeString, UWordBreak, utext_close, utext_extract,
    utext_native_length, utext_open_const_unicode_string,
)

# fmt: on


def test_break_iterator_adopt_text():
    where = Locale.get_english()
    builder = FilteredBreakIteratorBuilder.create_instance(where)
    assert isinstance(builder, FilteredBreakIteratorBuilder)
    fbi = builder.build(BreakIterator.create_sentence_instance(where))
    assert isinstance(fbi, BreakIterator)
    src = UnicodeString("hello world")
    it = StringCharacterIterator(src)

    # void icu::BreakIterator::adoptText(CharacterIterator *it)
    fbi.adopt_text(it)
    assert fbi.first() == 0
    assert fbi.next() == 11
    assert fbi.next() == BreakIterator.DONE

    fbi.adopt_text(None)
    assert fbi.first() == 0
    assert fbi.next() == BreakIterator.DONE


def test_break_iterator_api():
    where = Locale.get_english()
    builder = FilteredBreakIteratorBuilder.create_instance(where)
    assert isinstance(builder, FilteredBreakIteratorBuilder)
    fbi = builder.build(BreakIterator.create_word_instance(where))
    assert isinstance(fbi, BreakIterator)
    src = UnicodeString("hello world")
    it = StringCharacterIterator(src)
    fbi.adopt_text(it)

    # int32_t icu::BreakIterator::current(void)
    # int32_t icu::BreakIterator::first(void)
    # int32_t icu::BreakIterator::next(int32_t n)
    # int32_t icu::BreakIterator::next(void)
    assert fbi.first() == 0
    assert fbi.current() == 0
    assert fbi.next() == 5
    assert fbi.current() == 5
    assert fbi.next() == 6
    assert fbi.current() == 6
    assert fbi.next() == 11
    assert fbi.current() == 11
    assert fbi.next() == BreakIterator.DONE
    assert fbi.current() == 11

    assert fbi.first() == 0
    assert fbi.next(1) == 5
    assert fbi.next(1) == 6
    assert fbi.next(1) == 11
    assert fbi.next(-1) == 6

    # int32_t icu::BreakIterator::following(int32_t offset)
    assert fbi.following(0) == 5
    assert fbi.following(5) == 6
    assert fbi.following(6) == 11

    # Locale icu::BreakIterator::getLocale(
    #       ULocDataLocaleType type,
    #       UErrorCode &status
    # )
    loc = fbi.get_locale(ULocDataLocaleType.ULOC_VALID_LOCALE)
    assert isinstance(loc, Locale)
    assert loc == where

    loc = fbi.get_locale(ULocDataLocaleType.ULOC_ACTUAL_LOCALE)
    assert isinstance(loc, Locale)
    assert len(loc.get_name()) == 0

    # int32_t icu::BreakIterator::getRuleStatus()
    assert fbi.first() == 0
    assert fbi.get_rule_status() == UWordBreak.UBRK_WORD_NONE
    assert fbi.next() == 5
    assert fbi.get_rule_status() == UWordBreak.UBRK_WORD_NONE

    # int32_t icu::BreakIterator::getRuleStatusVec(
    #       int32_t *fillInVec,
    #       int32_t capacity,
    #       UErrorCode &status
    # )
    assert fbi.first() == 0
    fill_in_vec = fbi.get_rule_status_vec()
    assert isinstance(fill_in_vec, list)
    # assert len(fill_in_vec) == 0
    assert fill_in_vec == [0]  # correct?

    # UBool icu::BreakIterator::isBoundary(int32_t offset)
    assert fbi.is_boundary(0)
    assert not fbi.is_boundary(1)
    assert not fbi.is_boundary(2)
    assert not fbi.is_boundary(3)
    assert not fbi.is_boundary(4)
    assert fbi.is_boundary(5)
    assert fbi.is_boundary(6)
    assert not fbi.is_boundary(7)
    assert not fbi.is_boundary(8)
    assert not fbi.is_boundary(9)
    assert not fbi.is_boundary(10)
    assert fbi.is_boundary(11)

    # int32_t icu::BreakIterator::last(void)
    # int32_t icu::BreakIterator::previous(void)
    assert fbi.last() == 11
    assert fbi.current() == 11
    assert fbi.previous() == 6
    assert fbi.current() == 6
    assert fbi.previous() == 5
    assert fbi.current() == 5
    assert fbi.previous() == 0
    assert fbi.current() == 0

    # int32_t icu::BreakIterator::preceding(int32_t offset)
    assert fbi.preceding(12) == 11
    assert fbi.preceding(11) == 6
    assert fbi.preceding(6) == 5
    assert fbi.preceding(5) == 0

    # CharacterIterator &icu::BreakIterator::getText(void)
    it = fbi.get_text()
    assert isinstance(it, CharacterIterator)
    dest = UnicodeString()
    it.get_text(dest)
    assert dest == src

    # UText *icu::BreakIterator::getUText(
    #       UText *fillIn,
    #       UErrorCode &status
    # )
    ut = fbi.get_utext(None)
    dest = utext_extract(ut, 0, utext_native_length(ut))
    assert dest == src
    utext_close(ut)


def test_break_iterator_clone():
    where = Locale.get_english()
    builder = FilteredBreakIteratorBuilder.create_instance(where)
    fbi = builder.build(BreakIterator.create_sentence_instance(where))

    # BreakIterator *icu::BreakIterator::clone()
    fbi2 = fbi.clone()
    assert isinstance(fbi2, BreakIterator)


@pytest.mark.xfail(
    reason="FIXME: Filtered BreakIterator.__eq__(BreakIterator) is not work"
)
def test_break_iterator_eq():
    where = Locale.get_english()
    builder = FilteredBreakIteratorBuilder.create_instance(where)
    fbi = builder.build(BreakIterator.create_sentence_instance(where))
    text = UnicodeString("hello world")
    fbi.set_text(text)
    fbi2 = fbi.clone()

    # BreakIterator.__copy__() -> BreakIterator
    # BreakIterator.__deepcopy__(Optional[memo]) -> BreakIterator
    fbi2a = copy.copy(fbi)
    fbi2b = copy.deepcopy(fbi)
    # FIXME: Filtered BreakIterator.__eq__(BreakIterator) is not work.
    assert fbi == fbi2 == fbi2a == fbi2b


@pytest.mark.xfail(
    reason="FIXME: Filtered BreakIterator.__ne__(BreakIterator) is not work"
)
def test_break_iterator_ne():
    where = Locale.get_english()
    builder = FilteredBreakIteratorBuilder.create_instance(where)
    fbi = builder.build(BreakIterator.create_sentence_instance(where))
    text = UnicodeString("hello world")
    fbi.set_text(text)
    fbi2 = fbi.clone()

    # FIXME: Filtered BreakIterator.__ne__(BreakIterator) is not work.
    assert not (fbi != fbi2)


def test_filtered_break_iterator_builder_56():
    # [1]
    # static FilteredBreakIteratorBuilder *
    # icu::FilteredBreakIteratorBuilder::createInstance(
    #       const Locale &where,
    #       UErrorCode &status
    # )
    where = Locale.get_english()
    builder1 = FilteredBreakIteratorBuilder.create_instance(where)
    assert isinstance(builder1, FilteredBreakIteratorBuilder)

    builder1a = FilteredBreakIteratorBuilder.create_instance("en")
    assert isinstance(builder1a, FilteredBreakIteratorBuilder)

    # **Deprecated in ICU 60**
    # [2]
    # static FilteredBreakIteratorBuilder *
    # icu::FilteredBreakIteratorBuilder::createInstance(UErrorCode &status)
    builder2 = FilteredBreakIteratorBuilder.create_instance()
    assert isinstance(builder2, FilteredBreakIteratorBuilder)

    # UBool icu::FilteredBreakIteratorBuilder::suppressBreakAfter(
    #       const UnicodeString &string,
    #       UErrorCode &status
    # )
    assert builder2.suppress_break_after(UnicodeString("Mr."))
    assert builder2.suppress_break_after("Capt.")
    assert not builder2.suppress_break_after("Mr.")
    assert not builder2.suppress_break_after(UnicodeString("Capt."))

    # UBool icu::FilteredBreakIteratorBuilder::unsuppressBreakAfter(
    #       const UnicodeString &string,
    #       UErrorCode &status
    # )
    assert builder2.unsuppress_break_after(UnicodeString("Capt."))
    assert builder2.unsuppress_break_after("Mr.")
    assert not builder2.unsuppress_break_after("Capt.")
    assert not builder2.unsuppress_break_after(UnicodeString("Mr."))

    # **Deprecated in ICU 60**
    # BreakIterator *icu::FilteredBreakIteratorBuilder::build(
    #       BreakIterator *adoptBreakIterator,
    #       UErrorCode &status
    # )
    builder2.suppress_break_after("Mr.")
    fbi = builder2.build(BreakIterator.create_sentence_instance(where))
    assert isinstance(fbi, BreakIterator)

    # [1]
    # void icu::BreakIterator::setText(const UnicodeString &text)
    text = UnicodeString(
        "In the meantime Mr. Weston arrived with his small ship, which he had "
        "now recovered. Capt. Gorges, who informed the Sgt. here that one "
        "purpose of his going east was to meet with Mr. Weston, took this "
        "opportunity to call him to account for some abuses he had to lay to "
        "his charge."
    )
    fbi.set_text(text)
    assert fbi.next() == 84
    assert fbi.next() == 90
    assert fbi.next() == 278
    assert fbi.next() == BreakIterator.DONE

    text2 = UnicodeString()
    fbi.set_text(text2)
    assert fbi.next() == BreakIterator.DONE

    # [2]
    # void icu::BreakIterator::setText(
    #       UText *text,
    #       UErrorCode &status
    # )
    ut = utext_open_const_unicode_string(None, text)
    fbi.set_text(ut)
    assert fbi.next() == 84
    assert fbi.next() == 90
    assert fbi.next() == 278
    assert fbi.next() == BreakIterator.DONE
    utext_close(ut)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 60, reason="ICU4C<60")
def test_filtered_break_iterator_builder_60():
    # static FilteredBreakIteratorBuilder *
    # icu::FilteredBreakIteratorBuilder::createEmptyInstance(UErrorCode &status)
    builder = FilteredBreakIteratorBuilder.create_empty_instance()
    assert isinstance(builder, FilteredBreakIteratorBuilder)

    assert builder.suppress_break_after("Mr.")
    assert not builder.suppress_break_after("Mr.")
    assert builder.unsuppress_break_after("Mr.")
    assert not builder.unsuppress_break_after("Mr.")
    assert builder.suppress_break_after("Mr.")

    # BreakIterator *
    # icu::FilteredBreakIteratorBuilder::wrapIteratorWithFilter(
    #       BreakIterator *adoptBreakIterator,
    #       UErrorCode &status
    # )
    fbi = builder.wrap_iterator_with_filter(
        BreakIterator.create_sentence_instance(Locale.get_english())
    )
    assert isinstance(fbi, BreakIterator)

    text = UnicodeString(
        "In the meantime Mr. Weston arrived with his small ship, which he had "
        "now recovered. Capt. Gorges, who informed the Sgt. here that one "
        "purpose of his going east was to meet with Mr. Weston, took this "
        "opportunity to call him to account for some abuses he had to lay to "
        "his charge."
    )
    fbi.set_text(text)
    assert fbi.next() == 84
    assert fbi.next() == 90
    assert fbi.next() == 278
    assert fbi.next() == BreakIterator.DONE



import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 59:
    pytest.skip("ICU4C<59", allow_module_level=True)

# fmt: off
from icupy.icu import (
    U_EDITS_NO_RESET, U_FOLD_CASE_EXCLUDE_SPECIAL_I, U_OMIT_UNCHANGED_TEXT,
    U_TITLECASE_NO_BREAK_ADJUSTMENT, U_TITLECASE_NO_LOWERCASE, BreakIterator,
    CaseMap, Edits, UnicodeString,
)

# fmt: on


# From icu/source/test/intltest/strcase.cpp
def test_api():
    # static int32_t icu::CaseMap::fold(
    #       uint32_t options,
    #       const char16_t *src,
    #       int32_t srcLength,
    #       char16_t *dest,
    #       int32_t destCapacity,
    #       Edits *edits,
    #       UErrorCode &errorCode
    # )
    options = (
        U_OMIT_UNCHANGED_TEXT
        | U_EDITS_NO_RESET
        | U_FOLD_CASE_EXCLUDE_SPECIAL_I
    )
    src = "IßtanBul"
    edits = Edits()
    dest = CaseMap.fold(options, src, -1, edits)
    assert isinstance(dest, str)
    assert dest == "ıssb"

    dest = CaseMap.fold(options, src, -1, None)  # edits can be None
    assert isinstance(dest, str)
    assert dest == "ıssb"

    dest = CaseMap.fold(options, src, -1)  # edits is optional
    assert isinstance(dest, str)
    assert dest == "ıssb"

    dest = CaseMap.fold(options, src)  # src_length is optional
    assert isinstance(dest, str)
    assert dest == "ıssb"

    # static int32_t icu::CaseMap::toLower(
    #       const char *locale,
    #       uint32_t options,
    #       const char16_t *src,
    #       int32_t srcLength,
    #       char16_t *dest,
    #       int32_t destCapacity,
    #       Edits *edits,
    #       UErrorCode &errorCode
    # )
    locale = "tr"
    options = U_OMIT_UNCHANGED_TEXT
    src = "IstanBul"
    edits = Edits()
    dest = CaseMap.to_lower(locale, options, src, -1, edits)
    assert isinstance(dest, str)
    assert dest == "ıb"

    dest = CaseMap.to_lower(
        locale, options, src, -1, None
    )  # edits can be None
    assert isinstance(dest, str)
    assert dest == "ıb"

    dest = CaseMap.to_lower(locale, options, src, -1)  # edits is optional
    assert isinstance(dest, str)
    assert dest == "ıb"

    dest = CaseMap.to_lower(locale, options, src)  # src_length is optional
    assert isinstance(dest, str)
    assert dest == "ıb"

    # static int32_t icu::CaseMap::toTitle(
    #       const char *locale,
    #       uint32_t options,
    #       BreakIterator *iter,
    #       const char16_t *src,
    #       int32_t srcLength,
    #       char16_t *dest,
    #       int32_t destCapacity,
    #       Edits *edits,
    #       UErrorCode &errorCode
    # )
    locale = "nl"
    options = (
        U_OMIT_UNCHANGED_TEXT
        | U_TITLECASE_NO_BREAK_ADJUSTMENT
        | U_TITLECASE_NO_LOWERCASE
    )
    it = BreakIterator.create_word_instance(locale)
    src = "IjssEL IglOo"
    it.set_text(UnicodeString(src))
    it.first()
    edits = Edits()
    dest = CaseMap.to_title(locale, options, it, src, -1, edits)
    assert isinstance(dest, str)
    assert dest == "J"

    dest = CaseMap.to_title(
        locale, options, None, src, -1, None
    )  # edits can be None
    assert isinstance(dest, str)
    assert dest == "J"

    dest = CaseMap.to_title(
        locale, options, None, src, -1
    )  # edits is optional
    assert isinstance(dest, str)
    assert dest == "J"

    dest = CaseMap.to_title(
        locale, options, None, src
    )  # src_length is optional
    assert isinstance(dest, str)
    assert dest == "J"

    # static int32_t icu::CaseMap::toUpper(
    #       const char *locale,
    #       uint32_t options,
    #       const char16_t *src,
    #       int32_t srcLength,
    #       char16_t *dest,
    #       int32_t destCapacity,
    #       Edits *edits,
    #       UErrorCode &errorCode
    # )
    locale = "el"
    options = U_OMIT_UNCHANGED_TEXT
    src = "Πατάτα"
    edits = Edits()
    dest = CaseMap.to_upper(locale, options, src, -1, edits)
    assert isinstance(dest, str)
    assert dest == "ΑΤΑΤΑ"

    dest = CaseMap.to_upper(
        locale, options, src, -1, None
    )  # edits can be None
    assert isinstance(dest, str)
    assert dest == "ΑΤΑΤΑ"

    dest = CaseMap.to_upper(locale, options, src, -1)  # edits is optional
    assert isinstance(dest, str)
    assert dest == "ΑΤΑΤΑ"

    dest = CaseMap.to_upper(locale, options, src)  # src_length is optional
    assert isinstance(dest, str)
    assert dest == "ΑΤΑΤΑ"



import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 55:
    pytest.skip("ICU4C<55", allow_module_level=True)

import copy

# fmt: off
from icupy.icu import (
    Formattable, Locale, NumberFormat, ScientificNumberFormatter,
    UnicodeString,
)

# fmt: on


def test_clone():
    decfmt = NumberFormat.create_scientific_instance("en")
    decfmt.apply_pattern("0.00E+0")
    fmt1 = ScientificNumberFormatter.create_superscript_instance(decfmt)
    append_to = UnicodeString()

    # ScientificNumberFormatter *icu::ScientificNumberFormatter::clone()
    fmt2 = fmt1.clone()
    assert isinstance(fmt2, ScientificNumberFormatter)

    append_to.remove()
    fmt2.format(Formattable(6.02e23), append_to)
    assert append_to == "6.02\u00d710\u207a\u00b2\u00b3"

    fmt3 = copy.copy(fmt1)
    append_to.remove()
    fmt3.format(Formattable(6.02e23), append_to)
    assert append_to == "6.02\u00d710\u207a\u00b2\u00b3"

    fmt4 = copy.deepcopy(fmt1)
    append_to.remove()
    fmt4.format(Formattable(6.02e23), append_to)
    assert append_to == "6.02\u00d710\u207a\u00b2\u00b3"


def test_create_markup_instance():
    # [1]
    # static ScientificNumberFormatter *
    # icu::ScientificNumberFormatter::createMarkupInstance(
    #       const Locale &locale,
    #       const UnicodeString &beginMarkup,
    #       const UnicodeString &endMarkup,
    #       UErrorCode &status
    # )
    fmt = ScientificNumberFormatter.create_markup_instance(
        Locale.get_english(), UnicodeString("<sup>"), UnicodeString("</sup>")
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        "en", UnicodeString("<sup>"), UnicodeString("</sup>")
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        Locale.get_english(), "<sup>", UnicodeString("</sup>")
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        Locale.get_english(), UnicodeString("<sup>"), "</sup>"
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        "en", "<sup>", UnicodeString("</sup>")
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        "en", UnicodeString("<sup>"), "</sup>"
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        Locale.get_english(), "<sup>", "</sup>"
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        "en", "<sup>", "</sup>"
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    decfmt = NumberFormat.create_scientific_instance("en")
    decfmt.apply_pattern("0.00E+0")

    # [2]
    # static ScientificNumberFormatter *
    # icu::ScientificNumberFormatter::createMarkupInstance(
    #       DecimalFormat *fmtToAdopt,
    #       const UnicodeString &beginMarkup,
    #       const UnicodeString &endMarkup,
    #       UErrorCode &status
    # )
    fmt = ScientificNumberFormatter.create_markup_instance(
        decfmt, UnicodeString("<sup>"), UnicodeString("</sup>")
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        decfmt, "<sup>", UnicodeString("</sup>")
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        decfmt, UnicodeString("<sup>"), "</sup>"
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_markup_instance(
        decfmt, "<sup>", "</sup>"
    )
    assert isinstance(fmt, ScientificNumberFormatter)


def test_create_superscript_instance():
    # [1]
    # static ScientificNumberFormatter *
    # icu::ScientificNumberFormatter::createSuperscriptInstance(
    #       const Locale &locale,
    #       UErrorCode &status
    # )
    fmt = ScientificNumberFormatter.create_superscript_instance(
        Locale.get_english()
    )
    assert isinstance(fmt, ScientificNumberFormatter)

    fmt = ScientificNumberFormatter.create_superscript_instance("en")
    assert isinstance(fmt, ScientificNumberFormatter)

    decfmt = NumberFormat.create_scientific_instance("en")
    decfmt.apply_pattern("0.00E+0")

    # [2]
    # static ScientificNumberFormatter *
    # icu::ScientificNumberFormatter::createSuperscriptInstance(
    #       DecimalFormat *fmtToAdopt,
    #       UErrorCode &status
    # )
    fmt = ScientificNumberFormatter.create_superscript_instance(decfmt)
    assert isinstance(fmt, ScientificNumberFormatter)


def test_format():
    fmt = ScientificNumberFormatter.create_markup_instance(
        "en", "<sup>", "</sup>"
    )

    # UnicodeString &icu::ScientificNumberFormatter::format(
    #       const Formattable &number,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to = UnicodeString("String: ")
    result = fmt.format(Formattable(1.23456e-78), append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "String: 1.23456\u00d710<sup>-78</sup>"



import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 60:
    pytest.skip("ICU4C<60", allow_module_level=True)

import copy

# fmt: off
from icupy.icu import (
    INT32_MAX, CurrencyUnit, DecimalFormatSymbols, ErrorCode, Format,
    Formattable, Locale, MeasureUnit, UCurrencyUsage, UErrorCode,
    UnicodeString, UnicodeStringAppendable, UNumberDecimalSeparatorDisplay,
    UNumberFormatRoundingMode, UNumberSignDisplay, UNumberUnitWidth,
    UParseError,
)
from icupy.icu.number import (
    FormattedNumber, IntegerWidth, LocalizedNumberFormatter, Notation,
    NumberFormatter, ScientificNotation, UnlocalizedNumberFormatter,
)

# fmt: on


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_formatted_number_62():
    fmt = (
        NumberFormatter.with_()
        .notation(Notation.compact_short())
        .unit(MeasureUnit.get_kelvin())
        .locale("en-US")
    )
    num = fmt.format_double(65000)
    assert isinstance(num, FormattedNumber)

    # Appendable &icu::number::FormattedNumber::appendTo(
    #       Appendable &appendable,
    #       UErrorCode &status
    # )
    s = UnicodeString()
    appendable = UnicodeStringAppendable(s)
    result = num.append_to(appendable)
    assert isinstance(result, UnicodeStringAppendable)
    assert id(result) == id(appendable)
    assert s == "65K K"

    # UnicodeString icu::number::FormattedNumber::toString(UErrorCode &status)
    result = num.to_string()
    assert isinstance(result, UnicodeString)
    assert result == "65K K"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 64, reason="ICU4C<64")
def test_formatted_number_64():
    # fmt: off
    from icupy.icu import (
        ConstrainedFieldPosition, FormattedValue, UFieldCategory,
        UNumberFormatFields,
    )

    # fmt: on

    assert issubclass(FormattedNumber, FormattedValue)

    fmt = (
        NumberFormatter.with_()
        .notation(Notation.compact_short())
        .unit(MeasureUnit.get_kelvin())
        .locale("en-US")
    )
    num = fmt.format_double(65000)
    assert isinstance(num, FormattedNumber)

    # UBool icu::number::FormattedNumber::nextPosition(
    #       ConstrainedFieldPosition &cfpos,
    #       UErrorCode &status
    # )
    cfpos = ConstrainedFieldPosition()
    assert num.next_position(cfpos)
    assert cfpos.get_category() == UFieldCategory.UFIELD_CATEGORY_NUMBER
    assert cfpos.get_field() == UNumberFormatFields.UNUM_INTEGER_FIELD
    assert (cfpos.get_start(), cfpos.get_limit()) == (0, 2)

    assert num.next_position(cfpos)
    assert cfpos.get_category() == UFieldCategory.UFIELD_CATEGORY_NUMBER
    assert cfpos.get_field() == UNumberFormatFields.UNUM_COMPACT_FIELD
    assert (cfpos.get_start(), cfpos.get_limit()) == (2, 3)

    assert num.next_position(cfpos)
    assert cfpos.get_category() == UFieldCategory.UFIELD_CATEGORY_NUMBER
    assert cfpos.get_field() == UNumberFormatFields.UNUM_MEASURE_UNIT_FIELD
    assert (cfpos.get_start(), cfpos.get_limit()) == (4, 5)

    assert not num.next_position(cfpos)

    # UnicodeString icu::number::FormattedNumber::toTempString(
    #       UErrorCode &status
    # )
    result = num.to_temp_string()
    assert isinstance(result, UnicodeString)
    assert result == "65K K"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 65, reason="ICU4C<65")
def test_formatted_number_65():
    fmt = (
        NumberFormatter.with_()
        .notation(Notation.compact_short())
        .unit(MeasureUnit.get_kelvin())
        .locale("en-US")
    )
    num = fmt.format_double(65000)
    assert isinstance(num, FormattedNumber)

    # template<typename StringClass>
    # StringClass icu::number::FormattedNumber::toDecimalNumber(
    #       UErrorCode &status
    # )
    result = num.to_decimal_number()
    assert isinstance(result, str)
    assert result == "65"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 68, reason="ICU4C<68")
def test_formatted_number_68():
    fmt = (
        NumberFormatter.with_()
        .usage("person")
        .unit(MeasureUnit.get_kilogram())
        .locale("en-GB")
    )
    num = fmt.format_double(80)
    assert isinstance(num, FormattedNumber)

    # MeasureUnit icu::number::FormattedNumber::getOutputUnit(
    #       UErrorCode &status
    # )
    unit = num.get_output_unit()
    assert isinstance(unit, MeasureUnit)
    assert unit.get_identifier() == "stone-and-pound"


def test_integer_width():
    # icu::number::IntegerWidth
    assert isinstance(IntegerWidth.zero_fill_to(3), IntegerWidth)
    assert isinstance(
        IntegerWidth.zero_fill_to(3).truncate_at(3), IntegerWidth
    )


def test_localized_number_formatter_60():
    fmt = NumberFormatter.with_locale(Locale.get_us())
    assert isinstance(fmt, LocalizedNumberFormatter)

    # [2]
    # icu::number::LocalizedNumberFormatter::LocalizedNumberFormatter(
    #       const LocalizedNumberFormatter &other
    # )
    fmt2 = LocalizedNumberFormatter(fmt)

    # FormattedNumber icu::number::LocalizedNumberFormatter::formatDecimal(
    #       StringPiece value,
    #       UErrorCode &status
    # )
    num = fmt2.format_decimal("9.87E-3")
    assert isinstance(num, FormattedNumber)

    # FormattedNumber icu::number::LocalizedNumberFormatter::formatDouble(
    #       double value,
    #       UErrorCode &status
    # )
    num = fmt2.format_double(9.87e-3)
    assert isinstance(num, FormattedNumber)

    # FormattedNumber icu::number::LocalizedNumberFormatter::formatInt(
    #       int64_t value,
    #       UErrorCode &status
    # )
    num = fmt2.format_int(INT32_MAX + 1)
    assert isinstance(num, FormattedNumber)

    # UBool icu::number::NumberFormatterSettings<Derived>::copyErrorTo(
    #       UErrorCode &outErrorCode
    # )
    out_error_code = ErrorCode()
    assert not fmt2.copy_error_to(out_error_code)
    assert out_error_code.get() == UErrorCode.U_ZERO_ERROR

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::decimal(
    #       UNumberDecimalSeparatorDisplay style
    # )
    result = fmt2.decimal(
        UNumberDecimalSeparatorDisplay.UNUM_DECIMAL_SEPARATOR_AUTO
    )
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::integerWidth(
    #       const IntegerWidth &style
    # )
    result = fmt2.integer_width(IntegerWidth.zero_fill_to(2))
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::notation(
    #       const Notation &notation
    # )
    result = fmt2.notation(Notation.compact_long())
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::symbols(
    #       const DecimalFormatSymbols &symbols
    # )
    result = fmt.symbols(DecimalFormatSymbols(Locale.get_us()))
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::unit(
    #       const icu::MeasureUnit &unit
    # )
    result = fmt.unit(MeasureUnit.create_meter_per_second())
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::unitWidth(
    #       UNumberUnitWidth width
    # )
    result = fmt.unit_width(UNumberUnitWidth.UNUM_UNIT_WIDTH_FULL_NAME)
    assert isinstance(result, LocalizedNumberFormatter)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 61, reason="ICU4C<61")
def test_localized_number_formatter_61():
    from icupy.icu import UNumberGroupingStrategy

    fmt = NumberFormatter.with_locale(Locale.get_us())
    assert isinstance(fmt, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::grouping(
    #       UNumberGroupingStrategy strategy
    # )
    result = fmt.grouping(UNumberGroupingStrategy.UNUM_GROUPING_MIN2)
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::perUnit(
    #       const icu::MeasureUnit &perUnit
    # )
    result = fmt.per_unit(MeasureUnit.get_meter())
    assert isinstance(result, LocalizedNumberFormatter)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_localized_number_formatter_62():
    from icupy.icu.number import Precision, Scale

    # [1]
    # icu::number::LocalizedNumberFormatter::LocalizedNumberFormatter()
    fmt = LocalizedNumberFormatter()

    # Format *icu::number::LocalizedNumberFormatter::toFormat(
    #       UErrorCode &status
    # )
    fmt2 = (
        fmt.notation(Notation.engineering())
        .unit(MeasureUnit.get_kelvin())
        .to_format()
    )
    assert isinstance(fmt2, Format)

    obj = Formattable(65000)
    append_to = UnicodeString()
    fmt2.format(obj, append_to)
    assert append_to == "65E3 K"

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::precision(
    #       const Precision &precision
    # )
    result = fmt.precision(Precision.fixed_fraction(2))
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::roundingMode(
    #       UNumberFormatRoundingMode roundingMode
    # )
    result = fmt.rounding_mode(UNumberFormatRoundingMode.UNUM_ROUND_HALFEVEN)
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::scale(
    #       const Scale &scale
    # )
    result = fmt.scale(Scale.power_of_ten(2))
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::sign(
    #       UNumberSignDisplay style
    # )
    result = fmt.sign(UNumberSignDisplay.UNUM_SIGN_AUTO)
    assert isinstance(result, LocalizedNumberFormatter)

    # template<typename Derived>
    # UnicodeString icu::number::NumberFormatterSettings<Derived>::toSkeleton(
    #       UErrorCode &status
    # )
    result = fmt.unit(MeasureUnit.get_meter_per_second()).to_skeleton()
    assert isinstance(result, UnicodeString)
    assert result in (
        "unit/meter-per-second",
        "measure-unit/speed-meter-per-second",
    )


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 64, reason="ICU4C<64")
def test_localized_number_formatter_64():
    fmt = NumberFormatter.with_locale(Locale.get_us()).notation(
        Notation.engineering()
    )
    assert isinstance(fmt, LocalizedNumberFormatter)
    assert fmt.format_double(0.8765).to_string() == "876.5E-3"

    # template<typename Derived>
    # LocalPointer<Derived>
    # icu::number::NumberFormatterSettings<Derived>::clone()
    fmt2 = fmt.clone()
    assert isinstance(fmt2, LocalizedNumberFormatter)
    assert fmt2.format_double(0.8765).to_string() == "876.5E-3"

    fmt3 = copy.copy(fmt)
    assert fmt3.format_double(0.8765).to_string() == "876.5E-3"

    fmt4 = copy.deepcopy(fmt)
    assert fmt4.format_double(0.8765).to_string() == "876.5E-3"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 68, reason="ICU4C<68")
def test_localized_number_formatter_68():
    fmt = NumberFormatter.with_locale(Locale.get_us())
    assert isinstance(fmt, LocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::usage(
    #       StringPiece usage
    # )
    result = fmt.usage("person")
    assert isinstance(result, LocalizedNumberFormatter)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_localized_number_formatter_to_format():
    # fmt: off
    from icupy.icu import (
        FieldPosition, FieldPositionIterator, ICUError, ParsePosition,
    )

    # fmt: on
    # [1]
    # icu::number::LocalizedNumberFormatter::LocalizedNumberFormatter()
    fmt = LocalizedNumberFormatter()

    # Format *icu::number::LocalizedNumberFormatter::toFormat(
    #       UErrorCode &status
    # )
    fmt2 = (
        fmt.notation(Notation.engineering())
        .unit(MeasureUnit.get_kelvin())
        .to_format()
    )
    assert isinstance(fmt2, Format)

    # bool icu::Format::operator!=(const Format &other)
    fmt3 = fmt2.clone()
    assert fmt2 != fmt
    assert not (fmt2 != fmt3)

    # bool icu::Format::operator==(const Format &other)
    assert not (fmt2 == fmt)
    assert fmt2 == fmt3

    # [1]
    # UnicodeString &icu::Format::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPosition &pos,
    #       UErrorCode &status
    # )
    obj = Formattable(65000)
    append_to = UnicodeString()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt2.format(obj, append_to, pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert append_to == "65E3 K"

    # [2]
    # UnicodeString &icu::Format::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPositionIterator *posIter,
    #       UErrorCode &status
    # )
    obj = Formattable(65000)
    append_to.remove()
    pos_iter = FieldPositionIterator()
    result = fmt2.format(obj, append_to, pos_iter)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert append_to == "65E3 K"

    append_to.remove()
    pos_iter = None
    result = fmt2.format(obj, append_to, pos_iter)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert append_to == "65E3 K"

    # [3]
    # UnicodeString &icu::Format::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    obj = Formattable(65000)
    append_to.remove()
    result = fmt2.format(obj, append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert append_to == "65E3 K"

    # [1]
    # void icu::Format::parseObject(
    #       const UnicodeString &source,
    #       Formattable &result,
    #       ParsePosition &parse_pos
    # )
    source = UnicodeString("65E3 K")
    result = Formattable()
    parse_pos = ParsePosition()
    fmt2.parse_object(source, result, parse_pos)
    assert parse_pos.get_index() == 0
    assert parse_pos.get_error_index() == 0

    source = "65E3 K"
    result = Formattable()
    parse_pos = ParsePosition()
    fmt2.parse_object(source, result, parse_pos)
    assert parse_pos.get_index() == 0
    assert parse_pos.get_error_index() == 0

    # [2]
    # void icu::Format::parseObject(
    #       const UnicodeString &source,
    #       Formattable &result,
    #       UErrorCode &status
    # )
    source = UnicodeString("65E3 K")
    result = Formattable()
    with pytest.raises(ICUError) as exc_info:
        fmt2.parse_object(source, result)
    assert exc_info.value.args[0] == UErrorCode.U_INVALID_FORMAT_ERROR

    source = "65E3 K"
    result = Formattable()
    with pytest.raises(ICUError) as exc_info:
        fmt2.parse_object(source, result)
    assert exc_info.value.args[0] == UErrorCode.U_INVALID_FORMAT_ERROR


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 68, reason="ICU4C<68")
def test_no_unit():
    from icupy.icu import NoUnit

    fmt = NumberFormatter.with_locale(Locale.get_us())

    unit = NoUnit.base()
    assert isinstance(unit, MeasureUnit)
    assert fmt.unit(unit).format_int(1).to_string() == "1"

    unit = NoUnit.percent()
    assert isinstance(unit, MeasureUnit)
    assert fmt.unit(unit).format_int(1).to_string() == "1%"

    unit = NoUnit.permille()
    assert isinstance(unit, MeasureUnit)
    assert fmt.unit(unit).format_int(1).to_string() == "1\u2030"  # 1‰


def test_number_formatter_60():
    # static UnlocalizedNumberFormatter icu::number::NumberFormatter::with()
    fmt = NumberFormatter.with_()
    assert isinstance(fmt, UnlocalizedNumberFormatter)

    # static LocalizedNumberFormatter icu::number::NumberFormatter::withLocale(
    #       const Locale &locale
    # )
    fmt = NumberFormatter.with_locale(Locale("en_US"))
    assert isinstance(fmt, LocalizedNumberFormatter)

    fmt = NumberFormatter.with_locale("en_US")
    assert isinstance(fmt, LocalizedNumberFormatter)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_number_formatter_62():
    # [1]
    # static UnlocalizedNumberFormatter
    # icu::number::NumberFormatter::forSkeleton(
    #       const UnicodeString &skeleton,
    #       UErrorCode &status
    # )
    fmt = NumberFormatter.for_skeleton(
        UnicodeString("measure-unit/speed-meter-per-second")
    )
    assert isinstance(fmt, UnlocalizedNumberFormatter)

    fmt = NumberFormatter.for_skeleton("measure-unit/speed-meter-per-second")
    assert isinstance(fmt, UnlocalizedNumberFormatter)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 64, reason="ICU4C<64")
def test_number_formatter_64():
    # [2]
    # static UnlocalizedNumberFormatter
    # icu::number::NumberFormatter::forSkeleton(
    #       const UnicodeString &skeleton,
    #       UParseError &perror,
    #       UErrorCode &status
    # )
    perror = UParseError()
    fmt = NumberFormatter.for_skeleton(
        UnicodeString("measure-unit/speed-meter-per-second"), perror
    )
    assert isinstance(fmt, UnlocalizedNumberFormatter)

    perror = UParseError()
    fmt = NumberFormatter.for_skeleton(
        "measure-unit/speed-meter-per-second", perror
    )
    assert isinstance(fmt, UnlocalizedNumberFormatter)


def test_notation():
    # icu::number::Notation
    assert isinstance(Notation.compact_long(), Notation)
    assert isinstance(Notation.compact_short(), Notation)
    assert isinstance(Notation.engineering(), ScientificNotation)
    assert isinstance(Notation.scientific(), ScientificNotation)
    assert isinstance(Notation.simple(), Notation)

    # icu::number::ScientificNotation
    assert issubclass(ScientificNotation, Notation)

    assert isinstance(
        (
            Notation.engineering().with_exponent_sign_display(
                UNumberSignDisplay.UNUM_SIGN_ALWAYS
            )
        ),
        ScientificNotation,
    )
    assert isinstance(
        Notation.engineering().with_min_exponent_digits(2), ScientificNotation
    )


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_precision_60_62():
    # fmt: off
    from icupy.icu.number import (
        CurrencyPrecision, FractionPrecision, IncrementPrecision, Precision,
    )

    # fmt: on
    # icu::number::Precision
    assert isinstance(
        Precision.currency(UCurrencyUsage.UCURR_USAGE_STANDARD),
        CurrencyPrecision,
    )
    assert isinstance(Precision.fixed_fraction(3), FractionPrecision)
    assert isinstance(Precision.increment(0.5), IncrementPrecision)
    assert isinstance(Precision.integer(), FractionPrecision)
    assert isinstance(Precision.max_fraction(3), FractionPrecision)
    assert isinstance(Precision.min_fraction(1), FractionPrecision)
    assert isinstance(Precision.min_max_fraction(1, 3), FractionPrecision)
    assert isinstance(Precision.unlimited(), Precision)

    # icu::number::CurrencyPrecision
    assert issubclass(CurrencyPrecision, Precision)

    assert isinstance(
        (
            Precision.currency(
                UCurrencyUsage.UCURR_USAGE_STANDARD
            ).with_currency(CurrencyUnit("USD"))
        ),
        Precision,
    )

    # icu::number::FractionPrecision
    assert issubclass(FractionPrecision, Precision)

    assert isinstance(
        Precision.fixed_fraction(2).with_max_digits(2), Precision
    )
    assert isinstance(
        Precision.fixed_fraction(2).with_min_digits(3), Precision
    )

    # icu::number::IncrementPrecision
    assert issubclass(IncrementPrecision, Precision)

    assert isinstance(Precision.increment(0.5).with_min_fraction(1), Precision)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_precision_62():
    from icupy.icu.number import Precision

    # icu::number::Precision
    assert isinstance(Precision.fixed_significant_digits(3), Precision)
    assert isinstance(Precision.max_significant_digits(4), Precision)
    assert isinstance(Precision.min_max_significant_digits(3, 4), Precision)
    assert isinstance(Precision.min_significant_digits(3), Precision)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 69, reason="ICU4C<69")
def test_precision_69():
    from icupy.icu import UNumberRoundingPriority, UNumberTrailingZeroDisplay
    from icupy.icu.number import Precision

    # icu::number::Precision
    assert isinstance(
        (
            Precision.fixed_fraction(2).trailing_zero_display(
                UNumberTrailingZeroDisplay.UNUM_TRAILING_ZERO_AUTO
            )
        ),
        Precision,
    )

    # icu::number::FractionPrecision
    assert isinstance(
        (
            Precision.fixed_fraction(1).with_significant_digits(
                3, 3, UNumberRoundingPriority.UNUM_ROUNDING_PRIORITY_RELAXED
            )
        ),
        Precision,
    )


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_scale():
    from icupy.icu.number import Scale

    # icu::number::Scale
    assert isinstance(Scale.by_decimal("5.2"), Scale)
    assert isinstance(Scale.by_double(5.2), Scale)
    assert isinstance(Scale.by_double_and_power_of_ten(5.2, 3), Scale)
    assert isinstance(Scale.none(), Scale)
    assert isinstance(Scale.power_of_ten(2), Scale)


def test_unlocalized_number_formatter_60():
    fmt = NumberFormatter.with_()
    assert isinstance(fmt, UnlocalizedNumberFormatter)

    # [2]
    # icu::number::UnlocalizedNumberFormatter::UnlocalizedNumberFormatter(
    #       const UnlocalizedNumberFormatter &other
    # )
    fmt2 = UnlocalizedNumberFormatter(fmt)

    # LocalizedNumberFormatter icu::number::UnlocalizedNumberFormatter::locale(
    #       const icu::Locale &locale
    # )
    result = fmt2.locale(Locale.get_us())
    assert isinstance(result, LocalizedNumberFormatter)

    result = fmt2.locale("en-US")
    assert isinstance(result, LocalizedNumberFormatter)

    # UBool icu::number::NumberFormatterSettings<Derived>::copyErrorTo(
    #       UErrorCode &outErrorCode
    # )
    out_error_code = ErrorCode()
    assert not fmt2.copy_error_to(out_error_code)
    assert out_error_code.get() == UErrorCode.U_ZERO_ERROR

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::decimal(
    #       UNumberDecimalSeparatorDisplay style
    # )
    result = fmt2.decimal(
        UNumberDecimalSeparatorDisplay.UNUM_DECIMAL_SEPARATOR_AUTO
    )
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::integerWidth(
    #       const IntegerWidth &style
    # )
    result = fmt2.integer_width(IntegerWidth.zero_fill_to(2))
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::notation(
    #       const Notation &notation
    # )
    result = fmt2.notation(Notation.compact_long())
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::symbols(
    #       const DecimalFormatSymbols &symbols
    # )
    result = fmt.symbols(DecimalFormatSymbols(Locale.get_us()))
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::unit(
    #       const icu::MeasureUnit &unit
    # )
    result = fmt.unit(MeasureUnit.create_meter_per_second())
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::unitWidth(
    #       UNumberUnitWidth width
    # )
    result = fmt.unit_width(UNumberUnitWidth.UNUM_UNIT_WIDTH_FULL_NAME)
    assert isinstance(result, UnlocalizedNumberFormatter)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 61, reason="ICU4C<61")
def test_unlocalized_number_formatter_61():
    from icupy.icu import UNumberGroupingStrategy

    fmt = NumberFormatter.with_()
    assert isinstance(fmt, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::grouping(
    #       UNumberGroupingStrategy strategy
    # )
    result = fmt.grouping(UNumberGroupingStrategy.UNUM_GROUPING_MIN2)
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::perUnit(
    #       const icu::MeasureUnit &perUnit
    # )
    result = fmt.per_unit(MeasureUnit.get_meter())
    assert isinstance(result, UnlocalizedNumberFormatter)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 62, reason="ICU4C<62")
def test_unlocalized_number_formatter_62():
    from icupy.icu.number import Precision, Scale

    # [1]
    # icu::number::UnlocalizedNumberFormatter::UnlocalizedNumberFormatter()
    fmt = UnlocalizedNumberFormatter()

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::precision(
    #       const Precision &precision
    # )
    result = fmt.precision(Precision.fixed_fraction(2))
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::roundingMode(
    #       UNumberFormatRoundingMode roundingMode
    # )
    result = fmt.rounding_mode(UNumberFormatRoundingMode.UNUM_ROUND_HALFEVEN)
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::scale(
    #       const Scale &scale
    # )
    result = fmt.scale(Scale.power_of_ten(2))
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::sign(
    #       UNumberSignDisplay style
    # )
    result = fmt.sign(UNumberSignDisplay.UNUM_SIGN_AUTO)
    assert isinstance(result, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # UnicodeString icu::number::NumberFormatterSettings<Derived>::toSkeleton(
    #       UErrorCode &status
    # )
    result = fmt.unit(MeasureUnit.get_meter_per_second()).to_skeleton()
    assert isinstance(result, UnicodeString)
    assert result in (
        "unit/meter-per-second",
        "measure-unit/speed-meter-per-second",
    )


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 64, reason="ICU4C<64")
def test_unlocalized_number_formatter_64():
    fmt = NumberFormatter.with_().notation(Notation.engineering())
    assert isinstance(fmt, UnlocalizedNumberFormatter)
    assert fmt.locale("en-US").format_double(0.8765).to_string() == "876.5E-3"

    # template<typename Derived>
    # LocalPointer<Derived>
    # icu::number::NumberFormatterSettings<Derived>::clone()
    fmt2 = fmt.clone()
    assert isinstance(fmt2, UnlocalizedNumberFormatter)
    assert fmt2.locale("en-US").format_double(0.8765).to_string() == "876.5E-3"

    fmt3 = copy.copy(fmt)
    assert fmt3.locale("en-US").format_double(0.8765).to_string() == "876.5E-3"

    fmt4 = copy.deepcopy(fmt)
    assert fmt4.locale("en-US").format_double(0.8765).to_string() == "876.5E-3"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 68, reason="ICU4C<68")
def test_unlocalized_number_formatter_68():
    fmt = NumberFormatter.with_()
    assert isinstance(fmt, UnlocalizedNumberFormatter)

    # template<typename Derived>
    # Derived icu::number::NumberFormatterSettings<Derived>::usage(
    #       StringPiece usage
    # )
    result = fmt.usage("person")
    assert isinstance(result, UnlocalizedNumberFormatter)
